import numpy as np
import os
from PIL import Image, ImageDraw, ImageFont

def get_text_image(text, output_file, font_file, size):
    MAX_W, MAX_H = 12800,7200

    im = Image.new('RGBA', (MAX_W, MAX_H))
    font = ImageFont.truetype(
        font_file, size)

    relative_w, relative_h = font.getsize(text)

    mid_w, mid_h = MAX_W / 2, MAX_H / 2
    epsilon = 5
    im = im.crop((mid_w - relative_w / 2 - epsilon, mid_h - relative_h / 2 - epsilon, mid_w + relative_w / 2 + epsilon, mid_h + relative_h / 2 + epsilon))

    draw = ImageDraw.Draw(im)

    draw.text((0,0), text, font=font)

    im.save(output_file)

def generate_characters(characters, font_file, size, folder='sb_char'):
    if folder not in os.listdir():
        os.mkdir(folder)
    
    f_str = folder + "/{}.png"
    for chrs in characters:
        f_name = str(ord(chrs))
        get_text_image(chrs, f_str.format(f_name), font_file, size)

def get_color(RGB: tuple, fp: str):
    color = Image.new('RGB', (1280, 720), RGB)
    color.save(fp)

def gradient(RGB, fp):
    # Define width and height of image
    W, H = 720, 720

    # Create solid red image
    im = Image.new(mode='RGB', size=(W,H), color=RGB)

    # Create radial alpha/transparency layer. 255 in centre, 0 at edge
    Y = np.linspace(-1, 1, H)[None, :]*255
    X = np.linspace(-1, 1, W)[:, None]*255
    alpha = (X**2 + Y**2) ** 0.5
    alpha = np.clip(0,255,alpha)

    # Push that radial gradient transparency onto red image and save
    im.putalpha(Image.fromarray(alpha.astype(np.uint8)))
    im.save(fp)

def rectangle(fp, width, height):
    im = Image.new(mode='RGB', size=(width, height), color=(255,255,255))

    im.save(fp)

def randomglitcheffect(fp, width, height, RGB, chance):
    im = Image.new(mode='RGB', size=(width, height), color=RGB)
    alp = np.random.choice([0,1], p=[1 - chance, chance], size=(width, height)) * 255

    im.putalpha(Image.fromarray(alp.astype(np.uint8)))
    im.save(fp)

def white_ball(fp):
    W, H = 500, 500
    im = Image.new(mode='RGB', size=(W,H), color=(255,255,255))
    
    Y = np.linspace(-1, 1, H)[None, :]*255
    X = np.linspace(-1, 1, W)[:, None]*255
    al = (X ** 2 + Y ** 2) ** 0.5
    al[al<=255] = 255
    al[al > 255] = 0
    
    
    im.putalpha(Image.fromarray(al.astype(np.uint8)))
    im.save(fp)


from copy import deepcopy as copy
from enum import Enum
from typing import List, Optional, Type
from typing_extensions import Literal
from .actions import Action, Loop

DEFAULT_PATH = ""

class Position(Enum):
    BACKGROUND = 0
    FOREGROUND = 1
    OVERLAY = 2

class Sprite:
    def __init__(self, file_name):
        self.filename = file_name
        self.action = []
    
    def add_action(self, action: Type[Action]):
        self.action.append(action)
    
    def add_actions(self, actions: List[Type[Action]]):
        self.action += actions
    
    def render(self, pos: Literal[Position.BACKGROUND, Position.FOREGROUND, Position.OVERLAY]):
        string = {
            Position.BACKGROUND: "Background",
            Position.FOREGROUND: "Foreground",
            Position.OVERLAY: "Overlay"
        }
        text = f'Sprite,{string[pos]},Centre,"{self.filename}",320,240\n ' + "\n ".join([i.render() for i in self.action])
        return text + "\n"

class StoryBoard:
    EVENT_TEXT = "[Events]\n//Background and Video events\n"
    BACKGROUND_TEXT = "//Storyboard Layer 0 (Background)\n"
    FAILPASS_TEXT = "//Storyboard Layer 1 (Fail)\n//Storyboard Layer 2 (Pass)\n"
    FOREGROUND_TEXT = "//Storyboard Layer 3 (Foreground)\n"
    OVERLAY_TEXT = "//Storyboard Layer 4 (Overlay)\n"
    SOUND_SAMPLES = "//Storyboard Sound Samples"
    background_object = []
    foreground_object = []
    overlay_object = []
    def __init__(self, background_objects=[], foreground_objects=[], overlay_objects=[]) -> None:
        self.background_object = background_objects
        self.foreground_object = foreground_objects
        self.overlay_object = overlay_objects
    
    def render(self):
        text = self.EVENT_TEXT + \
        self.BACKGROUND_TEXT + "".join([i.render(Position.BACKGROUND) for i in self.background_object]) + \
        self.FAILPASS_TEXT + self.FOREGROUND_TEXT + "".join([i.render(Position.FOREGROUND) for i in self.foreground_object]) + \
        self.OVERLAY_TEXT + "".join([i.render(Position.OVERLAY) for i in self.overlay_object])
        return text + self.SOUND_SAMPLES

    @staticmethod
    def is_sprite(text):
        return "Sprite" in text
    
    @staticmethod
    def is_loop_action(text):
        return "L" in text
    
    @staticmethod
    def is_in_loop_action(text):
        return "  " in text
    
    @staticmethod
    def is_action(text):
        return text[0] == " "

    @staticmethod
    def parse_sprite(sprite_text: str):
        sprite_details = sprite_text.split(",")
        sprite_type = sprite_details[0]
        sprite_overlay = sprite_details[1]
        sprite_alignment = sprite_details[2]
        sprite_filename = sprite_details[3][1:-1]
        return Sprite(sprite_filename)

    @staticmethod
    def parse_action(action_text: str):
        action_text = action_text.replace(" ", "")
        action_details = action_text.split(",")
        action_type = action_details[0]
        action_easing = int(action_details[1])
        action_start = int(action_details[2])
        if action_details[3]:
            action_end = int(action_details[3])
        else:
            action_end = action_start
        action_params = []
        for dt in action_details[4:]:
            try:
                action_params.append(int(dt))
            except:
                try:
                    action_params.append(float(dt))
                except:
                    action_params.append(dt)

            
        return Action(action_type, action_easing, action_start, action_end, action_params)

    @staticmethod
    def parse_loop(loop_text: str):
        L, start_time, loop_count = loop_text.split(",")
        start_time = int(start_time)
        loop_count = int(loop_count)
        return Loop(start_time, loop_count, [])

    def from_osb(self, osb_fp: str):
        osb = open(osb_fp, 'r', encoding='utf-8')
        osb_text = osb.read().split("\n")
        current_line = -1
        current_sprite = None
        current_loop = None
        currently_in_loop = False
        for line in osb_text:
            if line in self.BACKGROUND_TEXT:
                current_line = Position.BACKGROUND
                continue
            elif line in self.FAILPASS_TEXT:
                current_line = -1
                continue
            elif line in self.FOREGROUND_TEXT:
                current_line = Position.FOREGROUND
                continue
            elif line in self.OVERLAY_TEXT:
                current_line = Position.OVERLAY
                continue

            if currently_in_loop and (not self.is_in_loop_action(line)):
                current_sprite.add_action(copy(current_loop))
                currently_in_loop = False
            
            if self.is_sprite(line):
                if current_sprite is not None:
                    if current_line == Position.BACKGROUND:
                        self.background_object.append(copy(current_sprite))
                    elif current_line == Position.FOREGROUND:
                        self.foreground_object.append(copy(current_sprite))
                    elif current_line == Position.OVERLAY:
                        self.overlay_object.append(copy(current_sprite))

                current_sprite = self.parse_sprite(line)
                continue
            
            if self.is_loop_action(line):
                current_loop = self.parse_loop(line)
                currently_in_loop = True
                continue

            if currently_in_loop:
                current_loop.actions.append(self.parse_action(line))
                continue

            if self.is_action(line):
                current_sprite.add_action(self.parse_action(line))

        osb.close()
        return self
    
    def merge(self, sb2):
        t = copy(self)
        t.background_object += sb2.background_object
        t.foreground_object += sb2.foreground_object
        t.overlay_object += sb2.overlay_object
        return t
    
    def add_sprite(self, position: str, sprite: Sprite):
        assert position.lower() in ["background", "foreground", "overlay"]
        if position.lower() == "background":
            self.background_object.append(sprite)
        elif position.lower() == "foreground":
            self.foreground_object.append(sprite)
        elif position.lower() == "overlay":
            self.overlay_object.append(sprite)

    def osb(self, osb_fp):
        with open(osb_fp, 'w+') as osb:
            osb.write(self.render())

def merge_sb(f1, f2):
    sb1 = StoryBoard().from_osb(f1)
    sb2 = StoryBoard().from_osb(f2)

    return sb1.merge(sb2)


from typing import List, Type
class Action:
    def __init__(self, event_type: str, easing: int, start_time: int, end_time: int, params: list) -> None:
        self.event_type = event_type
        self.easing = easing
        self.start_time = start_time
        self.end_time = end_time if end_time != None else ""
        self.params = params
    
    def __repr__(self):
        return self.render()
    
    def render(self):
        param_text = ""
        for param in self.params:
            param_text += ","
            if isinstance(param, str):
                param_text += param
            elif param != None:
                param_text += param.__repr__()
            else:
                continue
                
        return f"{self.event_type},{self.easing},{self.start_time},{self.end_time}{param_text}"
    
class Scale(Action):
    def __init__(self, easing: int, start_time: int, end_time:int, start_scale: float, end_scale: float) -> None:
        self.start_scale, self.end_scale = start_scale, end_scale
        super().__init__('S', easing, start_time, end_time, [start_scale, end_scale])

class Move(Action):
    def __init__(self, easing: int, start_time: int, end_time: int, pos_start: tuple, pos_end: tuple) -> None:
        self.x_start, self.y_start = pos_start
        self.x_end, self.y_end = pos_end
        super().__init__('M', easing, start_time, end_time, [self.x_start, self.y_start, self.x_end, self.y_end])

class Fade(Action):
    def __init__(self, easing: int, start_time: int, end_time: int, opacity_start: float, opacity_end: float) -> None:
        self.opacity_start = opacity_start
        self.opacity_end = opacity_end
        super().__init__('F', easing, start_time, end_time, [opacity_start, opacity_end])

class Rotate(Action):
    def __init__(self, easing: int, start_time: int, end_time: int, angle_start: float, angle_end: float) -> None:
        self.angle_start = angle_start
        self.angle_end = angle_end
        super().__init__('R', easing, start_time, end_time, [angle_start, angle_end])

class Color(Action):
    def __init__(self, easing: int, start_time: int, end_time: int, RGB1: tuple, RGB2: tuple):
        self.r1, self.g1, self.b1 = RGB1
        self.r2, self.g2, self.b2 = RGB2
        super().__init__('C', easing, start_time, end_time, [self.r1, self.g1, self.b1, self.r2, self.g2, self.b2])

class VectorScale(Action):
    def __init__(self, easing: int, start_time: int, end_time: int, start_size: tuple, end_size: tuple) -> None:
        self.start_x, self.start_y = start_size
        self.end_x, self.end_y = end_size
        super().__init__('V', easing, start_time, end_time, [self.start_x, self.start_y, self.end_x, self.end_y])

class MoveX(Action):
    def __init__(self, easing: int, start_time: int, end_time: int, x_start: float, x_end: float) -> None:
        self.x_start = x_start; self.x_end = x_end
        super().__init__('MX', easing, start_time, end_time, [x_start, x_end])

class MoveY(Action):
    def __init__(self, easing: int, start_time: int, end_time: int, y_start: float, y_end: float) -> None:
        self.y_start = y_start; self.x_end = y_end
        super().__init__('MY', easing, start_time, end_time, [y_start, y_end])

class Loop():
    def __init__(self, start_time: int, loop_count: int, actions: List[Type[Action]]) -> None:
        self.start_time = start_time
        self.loop_count = loop_count
        self.actions = actions
    
    def __repr__(self):
        loop_text = f"L,{self.start_time},{self.loop_count}\n  "
        loop_text += "\n  ".join([action.__repr__() for action in self.actions])
        return loop_text
    
    def render(self):
        return self.__repr__()


import os
import json
import numpy as np

class HitObject:
    def __init__(self, code):
        data = code.split(",")
        self.lane = int(data[0])
        self.samplesound = int(data[1])
        try:
            self.offset = int(data[2])
        except:
            self.offset = data[2]
        self.type = int(data[3])
        self.hitsound = int(data[4])
        self.release = -1
        if self.type == 128:
            try:
                self.release = int(data[-1].split(":")[0])
            except:
                self.release = data[-1].split(":")[0]

    def encode(self):
        a = "0:0:0:0:"
        if self.type == 128:
            a = str(self.release) + ":" + a
        return "{},{},{},{},{},{}".format(self.lane, self.samplesound, self.offset, self.type, self.hitsound, a)

class TimingPoint:
    def __init__(self, code):
        data = code.split(",")
        self.offset = float(data[0])
        self.isBPM = int(data[-2])
        if self.isBPM == 1:
            self.velocity = 60000 / float(data[1])
        else:
            self.velocity = -100 / float(data[1])
        self.timeSignature = int(data[2])
        self.hitsoundvolume = int(data[5])
        self.isKiai = int(data[-1])
    
    def switch(self, mainbpm):
        if self.isBPM == 1:
            self.velocity = self.velocity / mainbpm
            self.isBPM = 0
        else:
            self.velocity = mainbpm * self.velocity
            self.isBPM = 1
    
    def encode(self):
        vel = 60000 / self.velocity if self.isBPM == 1 else -100 / self.velocity
        return "{},{},{},1,0,{},{},{}".format(self.offset, vel, self.timeSignature, self.hitsoundvolume, self.isBPM, self.isKiai)

class osufile:
    def __init__(self, data: str):
        self.data = data.split("\n")
        self.General = None
        self.editor = None
        self.metadata = None
        self.difficulty = None
        self.TimingPoints = list()
        self.HitObjects = list()
        self.initialize_data()
        
    def initialize_data(self):
        self.parseGeneral()
        self.parseEditor()
        self.parseMetadata()
        self.parseDifficulty()
        self.parseTimingPoints()
        self.parseHitObjects()
    
    def parseGroup(self, keyword):
        start = self.data.index("[{}]".format(keyword)) + 1
        end = self.data[start:].index("") + start
        group = dict()
        for data in self.data[start:end]:
            d = data.split(":")
            try:
                group[d[0]] = int(d[1])
            except:
                try:
                    group[d[0]] = float(d[1])
                except:
                    group[d[0]] = d[1]
        return group

    def parseGeneral(self):
        self.General = self.parseGroup("General")
    
    def parseEditor(self):
        self.editor = self.parseGroup("Editor")
        if "Bookmarks" in self.editor.keys():
            self.editor["Bookmarks"] = [int(k) for k in self.editor["Bookmarks"].split(",")]
    
    def parseMetadata(self):
        self.metadata = self.parseGroup("Metadata")
        self.metadata["Tags"] = self.metadata["Tags"].split(" ")
    
    def parseDifficulty(self):
        self.difficulty = self.parseGroup("Difficulty")
    
    def parseTimingPoints(self):
        start = self.data.index("[TimingPoints]") + 1
        end = self.data[start:].index("") + start
        for data in self.data[start:end]:
            self.TimingPoints.append(TimingPoint(data))
    
    def parseHitObjects(self):
        start = self.data.index("[HitObjects]") + 1
        end = self.data[start:].index("") + start
        for data in self.data[start:end]:
            self.HitObjects.append(HitObject(data))
    
    def get_hitobjects(self, start, end):
        return list(filter(lambda x: start <= x.offset < end, self.HitObjects))
    
    def __getitem__(self, rnge: slice):
        t = self
        start = rnge.start
        stop = rnge.stop
        t.HitObjects = list(filter(lambda x: start <= x.offset < stop, t.HitObjects))
        return t
    
    def __repr__(self):
        initializestring = "osu file format v14\n\n"
        initializestring += "[General]\n"
        for key, value in zip(self.General.keys(), self.General.values()):
            initializestring += "{}:{}\n".format(key, value)
        initializestring += "\n"
        initializestring += "[Editor]\n"
        for key, value in zip(self.editor.keys(), self.editor.values()):
            initializestring += "{}:".format(key)
            if key == "Bookmarks":
                initializestring += str(self.editor[key])[1:len(str(self.editor[key])) - 1]
                initializestring += "\n"
                continue
            initializestring += "{}\n".format(value)
        initializestring += "\n"
        initializestring += "[Metadata]\n"
        for key, value in zip(self.metadata.keys(), self.metadata.values()):
            initializestring += "{}:".format(key)
            if key == "Tags":
                initializestring += " ".join(self.metadata[key])
                initializestring += "\n"
                continue
            initializestring += "{}\n".format(value)
        initializestring += "\n[Difficulty]\n"
        for key, value in zip(self.difficulty.keys(), self.difficulty.values()):
            initializestring += "{}:{}\n".format(key, value)
        initializestring += '\n[Events]\n//Background and Video events\n0,0,"blank.jpg",0,0\n//Break Periods\n//Storyboard Layer 0 (Background)\n//Storyboard Layer 1 (Fail)\n//Storyboard Layer 2 (Pass)\n//Storyboard Layer 3 (Foreground)\n//Storyboard Sound Samples\n'
        initializestring += "\n[TimingPoints]\n"
        for tp in self.TimingPoints:
            initializestring += tp.encode() + "\n"
        initializestring += "\n"
        initializestring += "[HitObjects]\n"
        for hit in self.HitObjects:
            initializestring += hit.encode() + "\n"
        initializestring += "\n"
        return initializestring


def parse_beatmap(file):
    # this doesn't work with japanese and unicode metadata
    with open(file, "r") as f:
        data = f.read()
        return osufile(data)


